/*
** File: coredma.c  
** Project: ADB3 core driver
** Purpose: Implements functions for DMA transfers in Core Interface.
**
** (C) Copyright Alpha Data 2009-2010
*/

#include <df.h>

#include "resource.h"
#include "coredma.h"

typedef enum _MainState {
  MainIdle = 0,
  MainInitial = 1,
  MainRunning = 2
} MainState;

typedef enum _SetupState {
  SetupIdle = 0,
  SetupRunning = 1,
} SetupState;

typedef enum _TransferState {
  TransferIdle = 0,
  TransferRunning = 1,
} TransferState;

typedef enum _TeardownState {
  TeardownIdle = 0,
  TeardownRunning = 1,
} TeardownState;

/* Forward declarations */

static void
dmaMapperCallback(
  DfDeviceObject* pDevObj,
  unsigned int mapIndex,
  DfDmaMapperNode* pTable,
  unsigned int tableLength,
  DfDmaMapperPosition* pProgress,
  void* pContext);

/* Low-level functions */

static void
dmaStartOnHardware(
  Adb3CoreDeviceContext* pDevCtx,
  unsigned int channelIndex)
{
  DmaChannelContext* pDmaContext = &pDevCtx->dma.channel[channelIndex];
  unsigned int transferIndex = pDmaContext->curr.state.transferIndex;

  pDmaContext->curr.state.transferIndex = transferIndex ^ 0x1;
  pDevCtx->methods.pDmaTransfer(
    pDevCtx,
    channelIndex,
    TRUE,
    pDmaContext->descriptor[transferIndex].aligned.buffer.pGeneric,
    pDmaContext->curr.table[transferIndex].length,
    pDmaContext->descriptor[transferIndex].aligned.busAddress);
}

static void
dmaStopOnHardware(
  Adb3CoreDeviceContext* pDevCtx,
  unsigned int channelIndex)
{
  pDevCtx->methods.pDmaTransfer(pDevCtx, channelIndex, FALSE, NULL, 0, 0);
}

static boolean_t
dmaSetup(
  Adb3CoreDeviceContext* pDevCtx,
  unsigned int channelIndex)
{
  DmaChannelContext* pDmaContext = &pDevCtx->dma.channel[channelIndex];
  unsigned int setupIndex = pDmaContext->curr.state.setupIndex;

  return dfDmaMapperSetupThread(
    pDevCtx->pDevObj,
    (channelIndex << 1) | setupIndex,
    pDmaContext->curr.parameters.pBuffer,
    dmaMapperCallback,
    pDevCtx,
    pDmaContext->curr.parameters.bWriteToLocal,
    pDmaContext->curr.parameters.bFixedLocalAddress,
    &pDmaContext->curr.slice,
    1,
    &pDmaContext->curr.position);
}

static void
dmaTeardown(
  Adb3CoreDeviceContext* pDevCtx,
  unsigned int channelIndex)
{
  DmaChannelContext* pDmaContext = &pDevCtx->dma.channel[channelIndex];
  unsigned int teardownIndex = pDmaContext->curr.state.teardownIndex;

  pDmaContext->curr.state.teardownIndex = teardownIndex ^ 0x1U;
  dfDmaMapperTeardown(pDevCtx->pDevObj, (channelIndex << 1) | teardownIndex);
}

static void
dmaCleanup(
  Adb3CoreDeviceContext* pDevCtx,
  unsigned int channelIndex)
{
  DmaChannelContext* pDmaContext = &pDevCtx->dma.channel[channelIndex];

  while (pDmaContext->curr.state.setupCount) {
    dmaTeardown(pDevCtx, channelIndex);
    pDmaContext->curr.state.setupCount--;
  }
  pDmaContext->curr.state.setupCount = 0;
  pDmaContext->curr.state.setupIndex = 0;
  pDmaContext->curr.state.transferCount = 0;
  pDmaContext->curr.state.transferIndex = 0;
  pDmaContext->curr.state.teardownCount = 0;
  pDmaContext->curr.state.teardownIndex = 0;
}

/* Callbacks */

static void
dmaMapperCallback(
  DfDeviceObject* pDevObj,
  unsigned int mapIndex,
  DfDmaMapperNode* pTable,
  unsigned int tableLength,
  DfDmaMapperPosition* pProgress,
  void* pContext)
{
  Adb3CoreDeviceContext* pDevCtx = (Adb3CoreDeviceContext*)pContext;
  unsigned int channelIndex = mapIndex >> 1;
  DmaChannelContext* pDmaContext = &pDevCtx->dma.channel[channelIndex];
  unsigned int setupIndex = pDmaContext->curr.state.setupIndex;
  DfSpinLockFlags f;

  dfDebugPrint(6, ("dmaMapperCallback: pDevCtx=0x%p mapIndex=%lu pTable=0x%p tableLength=%lu\n",
    (void*)pDevCtx, (unsigned long)mapIndex, (void*)pTable, (unsigned long)tableLength));

  pDmaContext->curr.table[setupIndex].pElements = pTable;
  pDmaContext->curr.table[setupIndex].length = tableLength;
  pDevCtx->methods.pDmaList(
    pDevCtx,
    pDmaContext->curr.parameters.bWriteToLocal,
    pDmaContext->curr.parameters.bFixedLocalAddress,
    pTable,
    tableLength,
    pDmaContext->descriptor[setupIndex].aligned.buffer.pGeneric,
    pDmaContext->descriptor[setupIndex].aligned.busAddress);
  pDmaContext->curr.state.setupIndex = setupIndex ^ 0x1;
  pDmaContext->curr.position = *pProgress;
  f = dfSpinLockGet(&pDmaContext->lock);
  pDmaContext->curr.state.setup = SetupIdle;
  pDmaContext->curr.state.setupCount++;
  dfAssert(pDmaContext->curr.state.setupCount <= 2);
  pDmaContext->curr.state.transferCount++;
  dfAssert(pDmaContext->curr.state.transferCount <= 2);
  while (1) {
    if (pDmaContext->curr.state.bCancel) {
      /* The DMA transfer is being cancelled */
      if (pDmaContext->curr.state.setup == SetupIdle && pDmaContext->curr.state.transfer == TransferIdle && pDmaContext->curr.state.teardown == TeardownIdle) {
        dfDebugPrint(1, ("dmaMapperCallback: cancelled, invoking callback, channelIndex=%lu\n", (unsigned long)channelIndex));
        pDmaContext->curr.state.main = MainIdle;
        pDmaContext->curr.state.bCancel = FALSE;
        dfSpinLockPut(&pDmaContext->lock, f);
        dmaCleanup(pDevCtx, channelIndex);
        pDmaContext->curr.pCallback(pDmaContext->curr.pTicket, pDmaContext->curr.pCallbackContext, CoreDmaCancelled);
        break;
      } else {
        dfDebugPrint(1, ("dmaMapperCallback: cancelled, waiting for all idle, channelIndex=%lu\n", (unsigned long)channelIndex));
        dfSpinLockPut(&pDmaContext->lock, f);
        break;
      }
    }
    if (pDmaContext->curr.state.transfer == TransferIdle && pDmaContext->curr.state.transferCount > 0) {
      /* Start DMA transfer on hardware */
      dfDebugPrint(6, ("dmaMapperCallback: starting DMA transfer on hardware, channelIndex=%lu\n", (unsigned long)channelIndex));
      pDmaContext->curr.state.transfer = TransferRunning;
      pDmaContext->curr.state.transferCount--;
      dfAssert(pDmaContext->curr.state.transferCount <= 2);
      dmaStartOnHardware(pDevCtx, channelIndex);
    } else {
      if (pDmaContext->curr.state.setup == SetupIdle && pDmaContext->curr.state.setupCount < 2 && pDmaContext->curr.position.sliceIndex != 1) {
        /* Set up DMA transfer */
        dfDebugPrint(6, ("dmaMapperCallback: setting up DMA transfer, channelIndex=%lu\n", (unsigned long)channelIndex));
        pDmaContext->curr.state.setup = SetupRunning;
        dfSpinLockPut(&pDmaContext->lock, f);
        dmaSetup(pDevCtx, channelIndex);
        f = dfSpinLockGet(&pDmaContext->lock);
      } else {
        if (pDmaContext->curr.state.teardown == TeardownIdle && pDmaContext->curr.state.teardownCount > 0) {
          /* Tear down DMA transfer */
          dfDebugPrint(6, ("dmaMapperCallback: tearing down DMA transfer, channelIndex=%lu\n", (unsigned long)channelIndex));
          pDmaContext->curr.state.teardown = TeardownRunning;
          dfSpinLockPut(&pDmaContext->lock, f);
          dmaTeardown(pDevCtx, channelIndex);
          f = dfSpinLockGet(&pDmaContext->lock);
          pDmaContext->curr.state.setupCount--;
          dfAssert(pDmaContext->curr.state.setupCount <= 2);
          pDmaContext->curr.state.teardownCount--;
          dfAssert(pDmaContext->curr.state.teardownCount <= 2);
          pDmaContext->curr.state.teardown = TeardownIdle;
          if (pDmaContext->curr.position.sliceIndex == 1 /* TO DO */ && pDmaContext->curr.state.setupCount == 0 && pDmaContext->curr.state.teardownCount == 0) {
            /* The DMA transfer is finished */
            dfDebugPrint(6, ("dmaMapperCallback: DMA transfer finished, channelIndex=%lu\n", (unsigned long)channelIndex));
            dfAssert(pDmaContext->curr.state.transferCount == 0);
            pDmaContext->curr.state.main = MainIdle;
            dfSpinLockPut(&pDmaContext->lock, f);
            pDmaContext->curr.pCallback(pDmaContext->curr.pTicket, pDmaContext->curr.pCallbackContext, CoreDmaSuccess);
            break;
          }
        } else {
          /* Nothing more to do */
          dfDebugPrint(6, ("dmaMapperCallback: nothing more to do, channelIndex=%lu\n", (unsigned long)channelIndex));
          dfSpinLockPut(&pDmaContext->lock, f);
          break;
        }
      }
    }
  }
}

void
dmaCompletedRoutine(
  DfDpc* pDpc,
  void* pContext, /* channel */
  void* pArg)
{
  Adb3CoreDeviceContext* pDevCtx = (Adb3CoreDeviceContext*)pContext;
  unsigned int channelIndex = (unsigned int)(uintptr_t)pArg;
  DmaChannelContext* pDmaContext = &pDevCtx->dma.channel[channelIndex];
  DfSpinLockFlags f;

  dfDebugPrint(6, ("dmaCompletedRoutine: pDevCtx=0x%p channelIndex=%lu\n", (void*)pDevCtx, (unsigned long)channelIndex));

  f = dfSpinLockGet(&pDmaContext->lock);
  if (pDmaContext->curr.state.transfer == TransferIdle) {
    dfSpinLockPut(&pDmaContext->lock, f);
    dfDebugPrint(0, ("+++ dmaCompletedRoutine: spurious call, pDevCtx=0x%p channelIndex=%lu\n", (void*)pDevCtx, (unsigned long)channelIndex));
    return;
  }
  pDmaContext->curr.state.transfer = TransferIdle;
  pDmaContext->curr.state.teardownCount++;
  dfAssert(pDmaContext->curr.state.teardownCount <= 2);
  while (1) {
    if (pDmaContext->curr.state.bCancel) {
      /* The DMA transfer is being cancelled */
      if (pDmaContext->curr.state.setup == SetupIdle && pDmaContext->curr.state.transfer == TransferIdle && pDmaContext->curr.state.teardown == TeardownIdle) {
        dfDebugPrint(1, ("dmaCompletedRoutine: cancelled, invoking callback, channelIndex=%lu\n", (unsigned long)channelIndex));
        pDmaContext->curr.state.main = MainIdle;
        pDmaContext->curr.state.bCancel = FALSE;
        dfSpinLockPut(&pDmaContext->lock, f);
        dmaCleanup(pDevCtx, channelIndex);
        pDmaContext->curr.pCallback(pDmaContext->curr.pTicket, pDmaContext->curr.pCallbackContext, CoreDmaCancelled);
        break;
      } else {
        dfDebugPrint(1, ("dmaCompletedRoutine: cancelled, waiting for all idle, channelIndex=%lu\n", (unsigned long)channelIndex));
        dfSpinLockPut(&pDmaContext->lock, f);
        break;
      }
    }
    if (pDmaContext->curr.state.transfer == TransferIdle && pDmaContext->curr.state.transferCount > 0) {
      /* Start DMA transfer on hardware */
      dfDebugPrint(6, ("dmaCompletedRoutine: starting DMA transfer on hardware, channelIndex=%lu\n", (unsigned long)channelIndex));
      pDmaContext->curr.state.transfer = TransferRunning;
      dmaStartOnHardware(pDevCtx, channelIndex);
      pDmaContext->curr.state.transferCount--;
      dfAssert(pDmaContext->curr.state.transferCount <= 2);
    } else {
      if (pDmaContext->curr.state.setup == SetupIdle && pDmaContext->curr.state.setupCount < 2 && pDmaContext->curr.position.sliceIndex != 1) {
        /* Setup DMA transfer */
        dfDebugPrint(6, ("dmaCompletedRoutine: setting up DMA transfer, channelIndex=%lu\n", (unsigned long)channelIndex));
        pDmaContext->curr.state.setup = SetupRunning;
        dfSpinLockPut(&pDmaContext->lock, f);
        dmaSetup(pDevCtx, channelIndex);
        f = dfSpinLockGet(&pDmaContext->lock);
      } else {
        if (pDmaContext->curr.state.teardown == TeardownIdle && pDmaContext->curr.state.teardownCount > 0) {
          /* Tear down DMA transfer */
          dfDebugPrint(6, ("dmaCompletedRoutine: tearing down DMA transfer, channelIndex=%lu\n", (unsigned long)channelIndex));
          pDmaContext->curr.state.teardown = TeardownRunning;
          dfSpinLockPut(&pDmaContext->lock, f);
          dmaTeardown(pDevCtx, channelIndex);
          f = dfSpinLockGet(&pDmaContext->lock);
          pDmaContext->curr.state.setupCount--;
          dfAssert(pDmaContext->curr.state.setupCount <= 2);
          pDmaContext->curr.state.teardownCount--;
          dfAssert(pDmaContext->curr.state.teardownCount <= 2);
          pDmaContext->curr.state.teardown = TeardownIdle;
          if (pDmaContext->curr.position.sliceIndex == 1 /* TO DO */ && pDmaContext->curr.state.setupCount == 0 && pDmaContext->curr.state.teardownCount == 0) {
            /* The DMA transfer is finished */
            dfDebugPrint(6, ("dmaCompletedRoutine: DMA transfer finished, channelIndex=%lu\n", (unsigned long)channelIndex));
            dfAssert(pDmaContext->curr.state.transferCount == 0);
            pDmaContext->curr.state.main = MainIdle;
            dfSpinLockPut(&pDmaContext->lock, f);
            pDmaContext->curr.pCallback(pDmaContext->curr.pTicket, pDmaContext->curr.pCallbackContext, CoreDmaSuccess);
            break;
          }
        } else {
          /* Nothing more to do */
          dfDebugPrint(6, ("dmaCompletedRoutine: nothing more to do, channelIndex=%lu\n", (unsigned long)channelIndex));
          dfSpinLockPut(&pDmaContext->lock, f);
          break;
        }
      }
    }
  }
}

/*
** Handle CoreDmaCancel
*/

CoreDmaStatus
coreDmaCancel(
  void* pInterfaceContext,
  CoreTicket* pTicket,
  unsigned int channelIndex)
{
  Adb3CoreDeviceContext* pDevCtx = (Adb3CoreDeviceContext*)pInterfaceContext;
  DmaChannelContext* pDmaContext = &pDevCtx->dma.channel[channelIndex];
  DfSpinLockFlags f;

  if (channelIndex >= pDevCtx->info.bootstrap.numDmaChannel) {
    return CoreDmaInvalidChannel;
  }

  if (!checkTicketValid(pDevCtx, pTicket, CORE_RESOURCE_DMAENGINE, channelIndex)) {
    return CoreDmaInvalidTicket;
  }

  f = dfSpinLockGet(&pDmaContext->lock);
  dfAssert(pDmaContext->curr.state.main != MainIdle);
  switch (pDmaContext->curr.state.main) {
  case MainInitial:
    dfDebugPrint(1, ("coreDmaCancel: in Initial state\n"));
    pDmaContext->curr.state.bCancel = TRUE;
    dfSpinLockPut(&pDmaContext->lock, f);
    break;

  case MainRunning:
    if (pDmaContext->curr.state.setup == SetupIdle && pDmaContext->curr.state.transfer == TransferIdle && pDmaContext->curr.state.teardown == TeardownIdle) {
      dfDebugPrint(1, ("coreDmaCancel: in Running state, cleaning up\n"));
      pDmaContext->curr.state.main = MainIdle;
      dfSpinLockPut(&pDmaContext->lock, f);
      dmaCleanup(pDevCtx, channelIndex);
      pDmaContext->curr.pCallback(pDmaContext->curr.pTicket, pDmaContext->curr.pCallbackContext, CoreDmaCancelled);
    } else {
      if (pDmaContext->curr.state.transfer == TransferRunning) {
        dfDebugPrint(1, ("coreDmaCancel: in Running state, aborting DMA on hardware\n"));
        dmaStopOnHardware(pDevCtx, channelIndex);
      } else {
        dfDebugPrint(1, ("coreDmaCancel: in Running state, not aborting DMA on hardware\n"));
      }
      pDmaContext->curr.state.bCancel = TRUE;
      dfSpinLockPut(&pDmaContext->lock, f);
    }
    break;
  }

  return TRUE;
}

/*
** Handle CoreDmaInitialize
*/

CoreDmaStatus
coreDmaInitialize(
  void* pInterfaceContext,
  CoreTicket* pTicket,
  unsigned int channelIndex,
  const CoreDmaParameters* pParameters,
  CoreDmaCallback* pCallback,
  void* pCallbackContext)
{
  Adb3CoreDeviceContext* pDevCtx = (Adb3CoreDeviceContext*)pInterfaceContext;
  DmaChannelContext* pDmaContext = &pDevCtx->dma.channel[channelIndex];

  if (channelIndex >= pDevCtx->info.bootstrap.numDmaChannel) {
    return CoreDmaInvalidChannel;
  }

  if (!checkTicketValid(pDevCtx, pTicket, CORE_RESOURCE_DMAENGINE, channelIndex)) {
    return CoreDmaInvalidTicket;
  }

  dfAssert(pDmaContext->curr.state.main != MainRunning);

  pDmaContext->curr.pTicket = pTicket;
  pDmaContext->curr.parameters = *pParameters;
  pDmaContext->curr.pCallback = pCallback;
  pDmaContext->curr.pCallbackContext = pCallbackContext;
  pDmaContext->curr.slice.length = pParameters->region.length;
  pDmaContext->curr.slice.offset = pParameters->region.offset;
  pDmaContext->curr.slice.localAddress = pParameters->region.localAddress;
  pDmaContext->curr.position.sliceIndex = 0;
  pDmaContext->curr.position.slicePosition = 0;

  pDmaContext->curr.state.bCancel = FALSE;
  pDmaContext->curr.state.main = MainInitial;
  pDmaContext->curr.state.setupIndex = 0;
  pDmaContext->curr.state.transferIndex = 0;
  pDmaContext->curr.state.teardownIndex = 0;

  return CoreDmaSuccess;
}

/*
** Handle CoreDmaTransfer
*/

void
coreDmaStart(
  void* pInterfaceContext,
  CoreTicket* pTicket,
  unsigned int channelIndex)
{
  Adb3CoreDeviceContext* pDevCtx = (Adb3CoreDeviceContext*)pInterfaceContext;
  DmaChannelContext* pDmaContext = &pDevCtx->dma.channel[channelIndex];
  DfSpinLockFlags f;

  if (channelIndex >= pDevCtx->info.bootstrap.numDmaChannel) {
    pDmaContext->curr.pCallback(pTicket, pDmaContext->curr.pCallbackContext, CoreDmaInvalidChannel);
  }

  if (!checkTicketValid(pDevCtx, pTicket, CORE_RESOURCE_DMAENGINE, channelIndex)) {
    pDmaContext->curr.pCallback(pTicket, pDmaContext->curr.pCallbackContext, CoreDmaInvalidTicket);
  }

  f = dfSpinLockGet(&pDmaContext->lock);
  dfAssert(pDmaContext->curr.state.main == MainInitial);
  dfAssert(pDmaContext->curr.state.setup == SetupIdle);
  dfAssert(pDmaContext->curr.state.setupCount == 0);
  dfAssert(pDmaContext->curr.state.transfer == TransferIdle);
  dfAssert(pDmaContext->curr.state.teardown == TeardownIdle);
  dfAssert(pDmaContext->curr.state.teardownCount == 0);
  if (pDmaContext->curr.state.bCancel) {
    /* We were cancelled */
    dfDebugPrint(1, ("coreDmaStart: cancelled, channelIndex=%lu\n", (unsigned long)channelIndex));
    pDmaContext->curr.state.bCancel = FALSE;
    pDmaContext->curr.state.main = MainIdle;
    dfSpinLockPut(&pDmaContext->lock, f);
    pDmaContext->curr.pCallback(pTicket, pDmaContext->curr.pCallbackContext, CoreDmaCancelled);
  } else {
    dfDebugPrint(6, ("coreDmaStart: starting setup, channelIndex=%lu\n", (unsigned long)channelIndex));
    pDmaContext->curr.state.main = MainRunning;
    pDmaContext->curr.state.setup = SetupRunning;
    dfSpinLockPut(&pDmaContext->lock, f);
    if (!dmaSetup(pDevCtx, channelIndex)) {
      f = dfSpinLockGet(&pDmaContext->lock);
      pDmaContext->curr.state.setup = SetupIdle;
      pDmaContext->curr.state.main = MainIdle;
      if (pDmaContext->curr.state.bCancel) {
        dfDebugPrint(1, ("coreDmaStart: setup failed, cancelled, channelIndex=%lu\n", (unsigned long)channelIndex));
        dfSpinLockPut(&pDmaContext->lock, f);
        pDmaContext->curr.pCallback(pDmaContext->curr.pTicket, pDmaContext->curr.pCallbackContext, CoreDmaCancelled);
      } else {
        dfDebugPrint(1, ("coreDmaStart: setup failed, channelIndex=%lu\n", (unsigned long)channelIndex));
        dfSpinLockPut(&pDmaContext->lock, f);
        pDmaContext->curr.pCallback(pTicket, pDmaContext->curr.pCallbackContext, CoreDmaUnexpectedError);
      }
    }
  }
}
