/*
** File: coreflash.c  
** Project: ADB3 core driver
** Purpose: Implements functions for Flash programming in Core Interface.
**
** (C) Copyright Alpha Data 2009-2010
*/

#include <df.h>

#include "coreflash.h"
#include "device.h"
#include "flash.h"
#include "resource.h"

CoreFlashStatus
coreFlashErase(
  void* pInterfaceContext,
  CoreTicket* pTicket,
  unsigned int flashIndex,
  uint64_t offset,
  uint64_t length)
{
  Adb3CoreDeviceContext* pDevCtx = (Adb3CoreDeviceContext*)pInterfaceContext;

  if (0 != flashIndex) {
    return CoreFlashInvalidIndex;
  }
  if (!checkTicketValid(pDevCtx, pTicket, CORE_RESOURCE_FLASH, flashIndex)) {
    return CoreClockInvalidTicket;
  }

  return flashErase(pDevCtx, flashIndex, offset, length);
}

CoreFlashStatus
coreGetFlashInfo(
  void* pInterfaceContext,
  unsigned int flashIndex,
  CoreFlashInfo* pFlashInfo)
{
  Adb3CoreDeviceContext* pDevCtx = (Adb3CoreDeviceContext*)pInterfaceContext;

  if (0 != flashIndex) {
    return CoreFlashInvalidIndex;
  }

  if (!pDevCtx->info.flash.bPresent) {
    return CoreFlashNotPresent;
  }

  pFlashInfo->size = pDevCtx->info.flash.totalSize;
  pFlashInfo->largestBlock = pDevCtx->info.flash.detail.validation.largestBlock;
  pFlashInfo->useableStart = pDevCtx->info.flash.targetArea.start;
  pFlashInfo->useableLength = pDevCtx->info.flash.targetArea.length;
  pFlashInfo->vendorId = pDevCtx->info.flash.detail.validation.vendorId;
  pFlashInfo->deviceId = pDevCtx->info.flash.detail.validation.deviceId;

  return CoreFlashSuccess;
}

CoreFlashStatus
coreGetFlashBlockInfo(
  void* pInterfaceContext,
  unsigned int flashIndex,
  uint64_t address,
  CoreFlashBlockInfo* pBlockInfo)
{
  Adb3CoreDeviceContext* pDevCtx = (Adb3CoreDeviceContext*)pInterfaceContext;
  CoreFlashStatus status;
  uint64_t blockStart, blockSize;

  status = flashIdentifyBlock(pDevCtx, flashIndex, address, &blockStart, &blockSize);
  if (status == CoreFlashSuccess) {
    pBlockInfo->address = blockStart;
    pBlockInfo->length = blockSize;
    pBlockInfo->flags = 0; /* TO DO */
  }
  return status;
}

CoreFlashStatus
coreFlashRead(
  void* pInterfaceContext,
  CoreTicket* pTicket,
  unsigned int flashIndex,
  uint64_t offset,
  size_t length,
  void* pBuffer)
{
  Adb3CoreDeviceContext* pDevCtx = (Adb3CoreDeviceContext*)pInterfaceContext;

  if (0 != flashIndex) {
    return CoreFlashInvalidIndex;
  }
  if (!checkTicketValid(pDevCtx, pTicket, CORE_RESOURCE_FLASH, flashIndex)) {
    return CoreClockInvalidTicket;
  }

  return flashRead(pDevCtx, flashIndex, offset, length, pBuffer);
}

CoreFlashStatus
coreFlashWrite(
  void* pInterfaceContext,
  CoreTicket* pTicket,
  unsigned int flashIndex,
  uint64_t offset,
  size_t length,
  const void* pBuffer)
{
  Adb3CoreDeviceContext* pDevCtx = (Adb3CoreDeviceContext*)pInterfaceContext;

  if (0 != flashIndex) {
    return CoreFlashInvalidIndex;
  }
  if (!checkTicketValid(pDevCtx, pTicket, CORE_RESOURCE_FLASH, flashIndex)) {
    return CoreClockInvalidTicket;
  }

  return flashWrite(pDevCtx, flashIndex, offset, length, pBuffer);
}
