/*
** File: icd2061.c  
** Project: ADM-XRC core driver
** Purpose: Functions for programming the ICD2061 clock generator.
**
** (C) Copyright Alpha Data 2008-2009
*/

#include <df.h>

#include "device.h"
#include "icd2061.h"

ClockWordStatus
clockWordIcd2061(
	Adb3CoreDeviceContext* pDevCtx,
  unsigned int clockIndex,
  uint32_t flags,
	uint32_t refFrequency,
	uint64_t frequencyReq,
	uint64_t frequencyMin,
	uint64_t frequencyMax,
  uint32_t* pClockWord,
	uint64_t* pActualFrequency)
{
  static const struct {
    uint32_t min, max;
  } mclkRange[16] = {
    { 49990000,  51000000 },
    { 50990000,  53200000 },
    { 53190000,  58500000 },
    { 58490000,  60700000 },
    { 60690000,  64400000 },
    { 60390000,  66800000 },
    { 66790000,  73500000 },
    { 73490000,  75600000 },
    { 75590000,  80900000 },
    { 80890000,  83200000 },
    { 83190000,  91500000 },
    { 91490000, 100000000 },
    { 99990000, 120100000 },
    { 99990000, 120100000 },
    { 99990000, 120100000 },
    { 99990000, 120100000 }
  };
  const uint32_t pMax = 131; /* Max allowable P, chosen to reduce jitter */
  uint32_t clockWord, i, d, p, q;
  uint32_t bestP = 0, bestQ = 0, bestD = 0; /* Init. to avoid compiler whine */
  uint32_t fVCO = 0;
  uint32_t error, bestError;
  uint32_t frequency, bestFrequency = 0; /* Init. to avoid compiler whine */
  uint32_t qMin, qMax; /* Min / max allowable Q */
  boolean_t bFound = FALSE;

  if (clockIndex >= 2) {
    return ClockWordInvalidIndex;
  }

  if (frequencyReq < 400000 || frequencyReq > 100000000 || clockIndex >= 2) {
    return ClockWordOutOfRange;
  }

  for (i = 0; i < 16; i++) {
    if ((fVCO > mclkRange[i].min) && (fVCO < mclkRange[i].max)) {
      break;
    }
  }

  /* Brute force P and Q guessing */
  bestError = 1000000000U;
  qMax = (refFrequency / 200000) - 1;
  qMin = (refFrequency / 5000000) + 1;
  for (d = 0; d < 8; d++) {
    for (q = qMin; q <= qMax; q++) {
      for (p = 4; p < pMax; p++) {
        fVCO = (uint32_t)refFrequency * 2 / q * p;
        if (fVCO < 50000000 || fVCO > 120000000) {
          continue;
        }
        /*
        ** The associativity of multiplication and divide operators is
        ** left to right. IF THIS FORMULA IS MODIFIED, ENSURE THAT IT
        ** CANNOT OVERFLOW AN (os_uint32).
        */
        frequency = fVCO / (1 << d);
        if ((flags & CoreClockMinimumFrequency) && frequency < frequencyMin) {
          continue;
        }
        if ((flags & CoreClockMaximumFrequency) && frequency > frequencyMax) {
          continue;
        }
        if (frequency > (uint32_t)frequencyReq) {
          error = frequency - (uint32_t)frequencyReq;
        } else {
          error = (uint32_t)frequencyReq - frequency;
        }
        if (error < bestError) {
          bestQ = q;
          bestP = p;
          bestD = d;
          bestError = error;
          bestFrequency = frequency;
          bFound = TRUE;
        }
      }
    }
  }

  if (!bFound) {
    return ClockWordGeneralFailure;
  }

  /* Assemble the completed word apart from the register selection */
  clockWord = (i << 17) | ((bestP - 3) << 10) | (bestD << 7) | (bestQ - 2);
	
  /*
  ** Add the ICD2061 register index
  **
  ** index = 0 means MCLK register
  ** index = 1 means VCLK1 register
  */
  switch (clockIndex) {
  case 0:
    clockWord |= (3 << 21);  /* ICD2061's MCLK */
    break;

  case 1:	
    clockWord |= (0 << 21);  /* ICD2061's VCLK1 */
    break;

  default:
    dfAssert(FALSE);
    break;
  }

  *pActualFrequency = bestFrequency;
  *pClockWord = clockWord;

  return ClockWordSuccess;
}
