/*
** File: model_adb3.c  
** Project: ADB3 core driver
** Purpose: ADB3 generic code, including function to bootstrap device context.
**
** (C) Copyright Alpha Data 2009-2010
*/

#include <df.h>

#include "adb3.h"
#include "adb3_common.h"
#include "device.h"
#include "model_boot.h"
#include "flash.h"

void
adb3DmaList(
  Adb3CoreDeviceContext* pDevCtx,
  boolean_t bWriteToDevice,
  boolean_t bFixedLocal,
  DfDmaMapperNode* pTable,
  unsigned int tableLength,
  void* pDescriptorBuffer,
  uint64_t descriptorBusAddress)
{
  Adb3DmaDescriptor* pDescriptor = (Adb3DmaDescriptor*)pDescriptorBuffer;
  uint64_t busAddress, nextBusAddress, tableBusAddress;
  unsigned int i;
  uint32_t ctl;

  dfDebugPrint(6, ("adb3DmaList: entered, pDevCtx=0x%p pTable=0x%p tableLength=%lu pDescriptorBuffer=0x%p\n",
    (void*)pDevCtx, (void*)pTable, (unsigned long)tableLength, pDescriptorBuffer));

  ctl = bWriteToDevice ? ADB3DMACTL_PCI_TO_LOCAL : ADB3DMACTL_LOCAL_TO_PCI;
  ctl |= bFixedLocal ? ADB3DMACTL_FIXED_LOCAL_ADDR : 0;
  busAddress = descriptorBusAddress;
  for (i = 0; i < tableLength; i++) {
    tableBusAddress = pTable->busAddress;
    nextBusAddress = busAddress + sizeof(Adb3DmaDescriptor);
    pDescriptor->pal = dfCpuToLe32((uint32_t)tableBusAddress);
    pDescriptor->pah = dfCpuToLe32((uint32_t)(tableBusAddress >> 32));
    pDescriptor->sizl = dfCpuToLe32((uint32_t)pTable->length);
    pDescriptor->sizh = dfCpuToLe32(0U);
    pDescriptor->ndl = dfCpuToLe32((uint32_t)nextBusAddress);
    pDescriptor->ndh = dfCpuToLe32((uint32_t)(nextBusAddress >> 32));
    pDescriptor->lal = dfCpuToLe32((uint32_t)pTable->localAddress);
    pDescriptor->ctl = dfCpuToLe32(ctl);
    pDescriptor++;
    pTable++;
    busAddress = nextBusAddress;
  }
  pDescriptor--;
  pDescriptor->ndl = dfCpuToLe32(0U);
  pDescriptor->ctl = dfCpuToLe32(ctl | ADB3DMACTL_ENDCHAIN | (bWriteToDevice ? ADB3DMACTL_IRQ_ON_FPGA_COMPLETION : ADB3DMACTL_IRQ_ON_HOST_COMPLETION));

#if DF_DBG_BUILD
  if (g_dfDebugLevel >= 7) {
    dfDebugPrint(0, ("adb3DmaList: dump of DMA engine linked list\n"));
    adb3DumpDmaDescriptors(pDevCtx, (Adb3DmaDescriptor*)pDescriptorBuffer, descriptorBusAddress, tableLength);
  }
#endif
}

void
adb3DmaTransfer(
  Adb3CoreDeviceContext* pDevCtx,
  unsigned channel,
  boolean_t bStart,
  void* pDescriptorBuffer,
  unsigned int numDescriptor,
  uint64_t descriptorBusAddress)
{
  Adb3DmaDescriptor* pDescriptor = (Adb3DmaDescriptor*)pDescriptorBuffer;
  Adb3GenericRegs* pRegs = pDevCtx->hardware.bridge.pAdb3;
  Adb3DmaRegs* pDmaRegs = &pRegs->dmaEngine[channel];
  uint32_t alignment, status;
  uint32_t lal, pal, ctl;

  dfDebugPrint(6, ("adb3DmaTransfer: entered, pDevCtx=0x%p channel=%lu bStart=%s pDescriptorBuffer=0x%p numDescriptor=%lu busAddress=0x%08lx_%08lx\n",
    (void*)pDevCtx, (unsigned long)channel, bStart ? "TRUE" : "FALSE", pDescriptorBuffer, (unsigned long)numDescriptor,
    dfSplitUint64(descriptorBusAddress)));

  if (bStart) {
    lal = pDescriptor->lal & 0xfU;
    pal = pDescriptor->pal & 0xfU;
    ctl = pDescriptor->ctl;
    if (ctl & ADB3DMACTL_PCI_TO_LOCAL) {
      if (pal > lal) {
        alignment = ((0x10U + lal - pal) & 0xfU) | ADB3DMAALIGN_LOCAL_IGNORE_FIRST_WORD;
      } else {
        alignment = (lal - pal) | ADB3DMAALIGN_LOCAL_USE_FIRST_WORD;
      }
      dfPciMemWrite32(&pDmaRegs->irqEnable, ADB3DMAIRQ_MPTL_WRITE_WORKER_COMPLETED | ADB3DMAIRQ_ABORT);
    } else {
      if (pal >= lal) {
        alignment = ((0x10U + lal - pal) & 0xf) | ADB3DMAALIGN_LOCAL_USE_FIRST_WORD;
      } else {
        alignment = (lal - pal) | ADB3DMAALIGN_LOCAL_IGNORE_FIRST_WORD;
      }
      dfPciMemWrite32(&pDmaRegs->irqEnable, ADB3DMAIRQ_PCIE_WRITE_WORKER_COMPLETED | ADB3DMAIRQ_ABORT);
    }
    dfPciMemWrite32(&pDmaRegs->alignment, alignment);
    dfPciMemWrite32(&pDmaRegs->nextAddrHigh, (uint32_t)(descriptorBusAddress >> 32));

#if DF_DBG_BUILD
    if (g_dfDebugLevel >= 7) {
      dfDebugPrint(0, ("adb3DmaTransfer: DMA channel %lu registers before starting on hardware:\n", (unsigned long)channel));
      adb3DumpDmaRegisters(pDevCtx, channel);
    }
#endif

    /* Writing to nextAddrLow starts the DMA transfer */
    dfPciMemWrite32(&pDmaRegs->nextAddrLow, (uint32_t)descriptorBusAddress); 

#if DF_DBG_BUILD
    if (g_dfDebugLevel >= 7) {
      dfDebugPrint(0, ("adb3DmaTransfer: DMA channel %lu registers after starting on hardware:\n", (unsigned long)channel));
      adb3DumpDmaRegisters(pDevCtx, channel);
    }
#endif
  } else {
#if DF_DBG_BUILD
    if (g_dfDebugLevel >= 7) {
      dfDebugPrint(0, ("adb3DmaTransfer: DMA channel %lu registers before attempt to abort:\n", (unsigned long)channel));
      adb3DumpDmaRegisters(pDevCtx, channel);
    }
#endif

    /* Write to ABORT register */ 
    dfPciMemWrite32(&pDmaRegs->abort,0);    
    /* Wait a bit for the bridge to halt the DMA */
    dfDelayBusyMicroseconds(15);
    /* Now check if DMA channel is active or not */
    status = dfPciMemRead32(&pDmaRegs->status);  
    if ((status & 0x3fffU) == 0x3fffU) {
      /* Everything idle, nothing to do */
    } else {
      /* Not in idle state, so assert cleanup strobe - should generate an interrupt */
      dfPciMemWrite32(&pDmaRegs->cleanup, 0);
      dfPciMemRead32(&pDmaRegs->cleanup);
    }

#if DF_DBG_BUILD
    if (g_dfDebugLevel >= 7) {
      dfDebugPrint(0, ("adb3DmaTransfer: DMA channel %lu registers after attempt to abort:\n", (unsigned long)channel));
      adb3DumpDmaRegisters(pDevCtx, channel);
    }
#endif
  }
}

void
adb3DumpDmaDescriptors(
	Adb3CoreDeviceContext* pDevCtx,
  Adb3DmaDescriptor* pDescriptors,
  uint64_t descBus,
  unsigned int count)
{
  Adb3DmaDescriptor* pCurrent;
  unsigned int i;

  if (sizeof(void*) == 4) {
    dfDebugPrint(0, ("  ### DescVirt   DescBus             [ PAH      PAL      SIZH     SIZL     NDH      NDL      LAL      CTL      ]\n"));
    for (i = 0; i < count; i++) {
      pCurrent = &pDescriptors[i];
      dfDebugPrint(0, ("  %03lu 0x%8p 0x%08lx_%08lx   %08lx %08lx %08lx %08lx %08lx %08lx %08lx %08lx\n",
        (unsigned long)i, (void*)pCurrent, dfSplitUint64(descBus),
        (unsigned long)dfLe32ToCpu(pCurrent->pah), (unsigned long)dfLe32ToCpu(pCurrent->pal),
        (unsigned long)dfLe32ToCpu(pCurrent->sizh), (unsigned long)dfLe32ToCpu(pCurrent->sizl),
        (unsigned long)dfLe32ToCpu(pCurrent->ndh), (unsigned long)dfLe32ToCpu(pCurrent->ndl),
        (unsigned long)dfLe32ToCpu(pCurrent->lal), (unsigned long)dfLe32ToCpu(pCurrent->ctl)));
      descBus += sizeof(Adb3DmaDescriptor);
    }
  } else {
    dfDebugPrint(0, ("  ### DescVirt          DescBus              [ PAH      PAL      SIZH     SIZL     NDH      NDL      LAL      CTL      ]\n"));
    for (i = 0; i < count; i++) {
      pCurrent = &pDescriptors[i];
      dfDebugPrint(0, ("  %03lu 0x%16p 0x%08lx_%08lx   %08lx %08lx %08lx %08lx %08lx %08lx %08lx %08lx\n",
        (unsigned long)i, (void*)pCurrent, dfSplitUint64(descBus),
        (unsigned long)dfLe32ToCpu(pCurrent->pah), (unsigned long)dfLe32ToCpu(pCurrent->pal),
        (unsigned long)dfLe32ToCpu(pCurrent->sizh), (unsigned long)dfLe32ToCpu(pCurrent->sizl),
        (unsigned long)dfLe32ToCpu(pCurrent->ndh), (unsigned long)dfLe32ToCpu(pCurrent->ndl),
        (unsigned long)dfLe32ToCpu(pCurrent->lal), (unsigned long)dfLe32ToCpu(pCurrent->ctl)));
      descBus += sizeof(Adb3DmaDescriptor);
    }
  }
}

void
adb3DumpDmaRegisters(
	Adb3CoreDeviceContext* pDevCtx,
  unsigned int channel)
{
  static const char* regNames[16] = {
    "status", "fifoStatus", "irqStatus", "irqEnable",
    "nextAddrLow", "nextAddrHigh", "hostControl", "fpgaControl",
    "alignment", "descPause", "reservedA", "reservedB",
    "reservedC", "reservedD", "reservedE", "reservedF"
  };
  uint32_t* pRegs = (uint32_t*)&pDevCtx->hardware.bridge.pAdb3->dmaEngine[channel];
  uint32_t regValue[16];
  unsigned int i;

  for (i = 0; i < 16; i++) {
    regValue[i] = dfPciMemRead32(pRegs + i);
  }
  for (i = 0; i < 4; i++) {
    dfDebugPrint(0, ("  %12s=0x%08lx %12s=0x%08lx %12s=0x%08lx %12s=0x%08lx\n",
      regNames[i], (unsigned long)regValue[i],
      regNames[i + 4], (unsigned long)regValue[i + 4],
      regNames[i + 8], (unsigned long)regValue[i + 8],
      regNames[i + 12], (unsigned long)regValue[i + 12]));
  }
}

static boolean_t
initHardware(
	Adb3CoreDeviceContext* pDevCtx,
	boolean_t bInit)
{
  Adb3GenericRegs* pRegs;

  dfDebugPrint(1, ("initHardware: bInit=%s\n", bInit ? "TRUE" : "FALSE"));

  pRegs = pDevCtx->hardware.bridge.pAdb3;

  if (bInit) {
    pDevCtx->info.model = CoreModelGeneric;

    /* Find base addresses and sizes of local bus windows */
    pDevCtx->info.window[0].busBase = pDevCtx->rawResources.variant.pci.memBar[2].base;
    pDevCtx->info.window[0].busSize = pDevCtx->rawResources.variant.pci.memBar[2].size;
    pDevCtx->info.window[0].localBase = 0;
    pDevCtx->info.window[0].localSize = pDevCtx->translatedResources.variant.pci.memBar[2].size;
    pDevCtx->info.window[0].translatedBase = pDevCtx->translatedResources.variant.pci.memBar[2].base;
    pDevCtx->info.window[0].translatedSize = pDevCtx->translatedResources.variant.pci.memBar[2].size;
    pDevCtx->info.window[0].pKernelBase = pDevCtx->hardware.pBridged;
    pDevCtx->info.window[0].kernelSize = (size_t)pDevCtx->translatedResources.variant.pci.memBar[2].size;
    pDevCtx->info.window[1].busBase = pDevCtx->rawResources.variant.pci.memBar[3].base;
    pDevCtx->info.window[1].busSize = pDevCtx->rawResources.variant.pci.memBar[3].size;
    pDevCtx->info.window[1].localBase = 0;
    pDevCtx->info.window[1].localSize = pDevCtx->translatedResources.variant.pci.memBar[3].size;
    pDevCtx->info.window[1].translatedBase = pDevCtx->translatedResources.variant.pci.memBar[3].base;
    pDevCtx->info.window[1].translatedSize = pDevCtx->translatedResources.variant.pci.memBar[3].size;
    pDevCtx->info.window[1].pKernelBase = pDevCtx->hardware.pBridgedPrefetch;
    pDevCtx->info.window[1].kernelSize = (size_t)pDevCtx->translatedResources.variant.pci.memBar[3].size;
    pDevCtx->info.window[2].busBase = pDevCtx->rawResources.variant.pci.memBar[0].base;
    pDevCtx->info.window[2].busSize = pDevCtx->rawResources.variant.pci.memBar[0].size;
    pDevCtx->info.window[2].localBase = 0;
    pDevCtx->info.window[2].localSize = 0;
    pDevCtx->info.window[2].translatedBase = pDevCtx->translatedResources.variant.pci.memBar[0].base;
    pDevCtx->info.window[2].translatedSize = pDevCtx->translatedResources.variant.pci.memBar[0].size;
    pDevCtx->info.window[2].pKernelBase = pDevCtx->hardware.bridge.pGeneric;
    pDevCtx->info.window[2].kernelSize = (size_t)pDevCtx->translatedResources.variant.pci.memBar[1].size;
    pDevCtx->info.window[3].busBase = pDevCtx->rawResources.variant.pci.memBar[1].base;
    pDevCtx->info.window[3].busSize = pDevCtx->rawResources.variant.pci.memBar[1].size;
    pDevCtx->info.window[3].localBase = 0;
    pDevCtx->info.window[3].localSize = 0;
    pDevCtx->info.window[3].translatedBase = pDevCtx->translatedResources.variant.pci.memBar[1].base;
    pDevCtx->info.window[3].translatedSize = pDevCtx->translatedResources.variant.pci.memBar[1].size;
    pDevCtx->info.window[3].pKernelBase = pDevCtx->hardware.model.pGeneric;
    pDevCtx->info.window[3].kernelSize = (size_t)pDevCtx->translatedResources.variant.pci.memBar[1].size;
  } else {
    /* Nothing to do */
  }

  return TRUE;
}

void
enableInterrupts(
	Adb3CoreDeviceContext* pDevCtx,
	boolean_t bEnable)
{
	Adb3GenericRegs* pRegs;
  
  pRegs = pDevCtx->hardware.bridge.pAdb3;

  dfDebugPrint(1, ("enableInterrupts: entered bEnable=%s\n", bEnable ? "TRUE" : "FALSE"));

	if (bEnable) {
		dfDebugPrint(1, ("enableInterrupts: interrupts enabled\n"));
	} else {
		dfDebugPrint(1, ("enableInterrupts: interrupts disabled\n"));
	}
}

boolean_t
isr(
  DfInterruptObject* pInterruptObj,
  void* pInterruptContext)
{
  Adb3CoreDeviceContext* pDevCtx = (Adb3CoreDeviceContext*)pInterruptContext;
  boolean_t bHandled = FALSE;
  Adb3GenericRegs* pRegs;

  dfDebugPrint(1, ("isr: entered\n"));

  pRegs = pDevCtx->hardware.bridge.pAdb3;
  
  return bHandled;
}

boolean_t
mapBusResources(
	Adb3CoreDeviceContext* pDevCtx,
	DfBusResources* pRaw,
	DfBusResources* pTranslated,
	boolean_t bUnmap)
{
	dfDebugPrint(2, ("mapBusResources: bUnmap=%s\n", bUnmap ? "TRUE" : "FALSE"));

	if (bUnmap) {
		if (NULL != pDevCtx->hardware.bridge.pGeneric) {
			if (!dfUnmapIoSpace(pDevCtx->pDevObj, pDevCtx->hardware.bridge.pGeneric)) {
				dfDebugPrint(0, ("*** mapBusResources: failed to unmap ADB3 generic registers\n"));
			}
      pDevCtx->hardware.bridge.pGeneric = NULL;
		}
		if (NULL != pDevCtx->hardware.model.pGeneric) {
			if (!dfUnmapIoSpace(pDevCtx->pDevObj, pDevCtx->hardware.model.pGeneric)) {
				dfDebugPrint(0, ("*** mapBusResources: failed to unmap ADB3 model-specific registers\n"));
			}
      pDevCtx->hardware.model.pGeneric = NULL;
		}
    if (NULL != pDevCtx->hardware.pBridged) {
			if (!dfUnmapIoSpace(pDevCtx->pDevObj, pDevCtx->hardware.pBridged)) {
				dfDebugPrint(0, ("*** mapBusResources: failed to unmap ADB3 bridged space\n"));
			}
      pDevCtx->hardware.pBridged = NULL;
		}
    if (NULL != pDevCtx->hardware.pBridgedPrefetch) {
			if (!dfUnmapIoSpace(pDevCtx->pDevObj, pDevCtx->hardware.pBridgedPrefetch)) {
				dfDebugPrint(0, ("*** mapBusResources: failed to unmap ADB3 prefetchable bridged space\n"));
			}
      pDevCtx->hardware.pBridgedPrefetch = NULL;
		}
		dfDebugPrint(2, ("mapBusResources: unmapped\n"));
	} else {
		pDevCtx->hardware.bridge.pGeneric = dfMapIoSpace(pDevCtx->pDevObj, pTranslated->variant.pci.memBar[0].base, 0x1000U);
		if (NULL == pDevCtx->hardware.bridge.pGeneric) {
			dfDebugPrint(0, ("*** mapBusResources: failed to map ADB3 generic registers\n"));
			return FALSE;
		} else {
			dfDebugPrint(2, ("mapBusResources: mapped ADB3 generic registers @ 0x%08lx_%08lx to 0x%p\n",
        dfSplitUint64(pTranslated->variant.pci.memBar[0].base), pDevCtx->hardware.bridge.pGeneric));
		}
		pDevCtx->hardware.model.pGeneric = dfMapIoSpace(pDevCtx->pDevObj, pTranslated->variant.pci.memBar[1].base, 0x1000U);
		if (NULL == pDevCtx->hardware.model.pGeneric) {
			dfDebugPrint(0, ("*** mapBusResources: failed to map ADB3 model-specific registers\n"));
			return FALSE;
		} else {
			dfDebugPrint(2, ("mapBusResources: mapped ADB3 model-specific registers @ 0x%08lx_%08lx to 0x%p\n",
        dfSplitUint64(pTranslated->variant.pci.memBar[1].base), pDevCtx->hardware.model.pGeneric));
		}
    pDevCtx->hardware.pBridged = dfMapIoSpace(pDevCtx->pDevObj, pTranslated->variant.pci.memBar[1].base, pTranslated->variant.pci.memBar[2].size);
		if (NULL == pDevCtx->hardware.pBridged) {
			dfDebugPrint(0, ("*** mapBusResources: failed to map ADB3 bridged space\n"));
			return FALSE;
		} else {
			dfDebugPrint(2, ("mapBusResources: mapped ADB3 bridged space @ 0x%08lx_%08lx to 0x%p\n",
        dfSplitUint64(pTranslated->variant.pci.memBar[2].base), pDevCtx->hardware.pBridged));
		}
    pDevCtx->hardware.pBridgedPrefetch = dfMapIoSpace(pDevCtx->pDevObj, pTranslated->variant.pci.memBar[1].base, pTranslated->variant.pci.memBar[3].size);
		if (NULL == pDevCtx->hardware.pBridgedPrefetch) {
			dfDebugPrint(0, ("*** mapBusResources: failed to map ADB3 prefetchable bridged space\n"));
			return FALSE;
		} else {
			dfDebugPrint(2, ("mapBusResources: mapped ADB3 prefetchable bridged space @ 0x%08lx_%08lx to 0x%p\n",
        dfSplitUint64(pTranslated->variant.pci.memBar[3].base), pDevCtx->hardware.pBridgedPrefetch));
		}
		dfDebugPrint(2, ("mapBusResources: mapped OK\n"));
	}

	return TRUE;
}

boolean_t
validateBusResources(
	Adb3CoreDeviceContext* pDevCtx,
	DfBusResources* pRaw,
	DfBusResources* pTranslated)
{
	if (pTranslated->variant.pci.numMemBar < 4) {
		dfDebugPrint(0, ("*** validateBusResources: less than 4 memory BARs\n"));
		return FALSE;
	}
	if (pTranslated->variant.pci.memBar[0].size < 0x1000U) {
		dfDebugPrint(0, ("*** validateBusResources: BAR0 smaller than 0x1000 bytes\n"));
		return FALSE;
	}
	if (pTranslated->variant.pci.memBar[1].size < 0x1000U) {
		dfDebugPrint(0, ("*** validateBusResources: BAR1 smaller than 0x1000 bytes\n"));
		return FALSE;
	}
	if (pTranslated->variant.pci.memBar[2].size < 0x100000U) {
		dfDebugPrint(0, ("*** validateBusResources: BAR2 smaller than 0x100000 bytes\n"));
		return FALSE;
	}
	if (pTranslated->variant.pci.memBar[3].size < 0x100000U) {
		dfDebugPrint(0, ("*** validateBusResources: BAR2 smaller than 0x100000 bytes\n"));
		return FALSE;
	}
	if (pTranslated->variant.pci.numIrq != 1) {
		dfDebugPrint(0, ("*** validateBusResources: not exactly 1 interrupt\n"));
		return FALSE;
	}

	dfDebugPrint(2, ("validateBusResources: OK\n"));

	return TRUE;
}

void
adb3CoreBootstrapGenericAdb3(
	Adb3CoreDeviceContext* pDevCtx)
{
  dfDebugPrint(1, ("adb3CoreBootstrapGenericAdb3: entered\n"));

  pDevCtx->methods.pEnableInterrupts = enableInterrupts;
  pDevCtx->methods.pInitHardware = initHardware;
  pDevCtx->methods.pIsr = isr;
  pDevCtx->methods.pMapBusResources = mapBusResources;
  pDevCtx->methods.pValidateBusResources = validateBusResources;

  pDevCtx->info.bootstrap.numClockGen = 0;
  pDevCtx->info.bootstrap.numDmaChannel = 1;
  pDevCtx->info.bootstrap.numTargetFpga = 0;
  pDevCtx->info.bootstrap.numWindow = 4;
  pDevCtx->info.bootstrap.bVpdInFlash = TRUE;
  pDevCtx->info.bootstrap.vpdLength = 0;
  pDevCtx->info.bootstrap.vpdOffset = 0;
  pDevCtx->info.bootstrap.flashDeviceWidth = 0;
  pDevCtx->info.bootstrap.flashWidth = 0;
  pDevCtx->info.bootstrap.bFlashStackedDie = FALSE;
  pDevCtx->info.bootstrap.b64BitDmaAddress = FALSE; /* TO DO - disable 64 bit PCIe addresses because not tested by AM/AC yet */
  pDevCtx->info.bootstrap.dmaDescriptorSize = 32;
}
