/*
** File: dma.c  
** Project: Linux Driver Framework
** Purpose: Implementation of DMA mapper functions
**
** (C) Copyright Alpha Data 2010
*/

#include "df.h"
#include "dflinux.h"

void*
dfDmaBufferAllocate(
  DfDeviceObject* pDevObj,
  size_t size,
  uint64_t* pBusAddress)
{
  void* pKernelAddress;
  dma_addr_t busAddress;

  dfAssert(pDevObj->hardwareId.busType == DfBusPci);
  if (size > dfDmaBufferMaxLength) {
    dfDebugPrint(0, ("*** dfDmaBufferAllocate: attempt to allocate more than %lu(0x%lx) bytes, pDevObj=0x%p size=%lu(0x%lx)\n",
		     (unsigned long)dfDmaBufferMaxLength, (unsigned long)dfDmaBufferMaxLength, (void*)pDevObj,
		     (unsigned long)size, (unsigned long)size));
    return NULL;
  }
  pKernelAddress = dma_alloc_coherent(&pDevObj->system.device.pPci->dev, size, &busAddress, GFP_KERNEL);
  if (NULL == pKernelAddress) {
    return NULL;
  }
  *pBusAddress = (uint64_t)busAddress;
  return pKernelAddress;
}

void
dfDmaBufferFree(
  DfDeviceObject* pDevObj,
  void* pKernelAddress,
  size_t size,
  uint64_t busAddress)
{
  dfAssert(pDevObj->hardwareId.busType == DfBusPci);
  dma_free_coherent(&pDevObj->system.device.pPci->dev, size, pKernelAddress, busAddress);
}

boolean_t
dfDmaMapperAllocate(
  DfDeviceObject* pDevObj,
  unsigned int numberOfMaps,
  unsigned int mapLength,          /* In blocks */
  boolean_t b64BitAddressing,
  unsigned int* pActualMapLength)  /* In blocks */
{
  unsigned int i, numMapRegister;
  DfLinuxDmaMapperContext* pMapArray = NULL;

  dfAssert(pDevObj->hardwareId.busType == DfBusPci);

  if (pDevObj->system.dma.pMapArray != NULL) {
    dfDebugPrint(0, ("*** dfDmaMapperAllocate: attempt to allocate mapper twice, pDevObj=0x%p\n", pDevObj));
    return FALSE;
  }

  pMapArray = (DfLinuxDmaMapperContext*)dfMalloc(numberOfMaps * sizeof(DfLinuxDmaMapperContext));
  if (NULL == pMapArray) {
    dfDebugPrint(0, ("*** dfDmaMapperAllocate: failed to allocate mapper context array, pDevObj=0x%p\n", pDevObj));
    return FALSE;
  }
  pDevObj->system.dma.pMapArray = pMapArray;
  pDevObj->system.dma.numMap = numberOfMaps;
  for (i = 0; i < numberOfMaps; i++) {
    pMapArray[i].pDeviceObject = pDevObj;
    pMapArray[i].index = i;
    pMapArray[i].pTable = NULL;
#if DF_DBG_BUILD
    pMapArray[i].curr.bIdle = TRUE;
#endif
  }

  numMapRegister = DF_DMA_MAPPER_MAX_TABLE_LENGTH;
  pDevObj->system.dma.maxMapRegister = numMapRegister;

  for (i = 0; i < numberOfMaps; i++) {
    pMapArray[i].pTable = (DfDmaMapperNode*)dfMalloc(numMapRegister * sizeof(DfDmaMapperNode));
    if (NULL == pMapArray[i].pTable) {
      dfDebugPrint(0, ("*** dfDmaMapperAllocate: failed to allocate table %lu, pDevObj=0x%p numberOfMaps=%lu mapLength=%lu nActualMapLength=%lu",
		       (unsigned long)i, pDevObj,
		       (unsigned long)numberOfMaps, (unsigned long)mapLength, (unsigned long)numMapRegister));
      goto failed;
    }
  }

  *pActualMapLength = numMapRegister;
  dfDebugPrint(3, ("dfDmaMapperAllocate: pDevObj=0x%p numberOfMaps=%lu mapLength=%lu nActualMapLength=%lu\n",
		   pDevObj, (unsigned long)numberOfMaps, (unsigned long)mapLength, (unsigned long)numMapRegister));
  return TRUE;

 failed:
  if (NULL != pMapArray) {
    for (i = 0; i < numberOfMaps; i++) {
      if (NULL != pMapArray[i].pTable) {
        dfFree(pMapArray[i].pTable);
        pMapArray[i].pTable = NULL;
      }
    }
    dfFree(pMapArray);
    pDevObj->system.dma.pMapArray = NULL;
  }
  return FALSE;
}

extern void
dfDmaMapperFree(
  DfDeviceObject* pDevObj)
{
  DfLinuxDmaMapperContext* pMapArray;
  unsigned int i, numberOfMaps;

  dfAssert(pDevObj->hardwareId.busType == DfBusPci);

  pMapArray = pDevObj->system.dma.pMapArray;
  numberOfMaps = pDevObj->system.dma.numMap;
  if (NULL != pMapArray) {
    for (i = 0; i < numberOfMaps; i++) {
      if (NULL != pMapArray[i].pTable) {
        dfFree(pMapArray[i].pTable);
        pMapArray[i].pTable = NULL;
      }
    }
    dfFree(pMapArray);
    pDevObj->system.dma.pMapArray = NULL;
  } else {
    dfDebugPrint(0, ("*** dfDmaMapperFree: attempting to free mapper when mapper not allocated, pDevObj=0x%p\n", pDevObj));
  }
}

extern boolean_t
dfDmaMapperSetup(
  DfDeviceObject* pDevObj,
  unsigned int mapIndex,
  DfBufferDescription* pBufferDescription,
  DfDmaMapperCallback* pCallback,
  void* pCallbackContext,
  boolean_t bWriteToDevice,
  boolean_t bFixedLocalAddress,
  DfDmaMapperSlice* pSlices,
  size_t numSlice,
  DfDmaMapperPosition* pStartPosition)
{
  const unsigned int pageOffsetMask = ~PAGE_MASK;
  struct device* pDev;
  DfDmaMapperSlice* pSlice;
  DfDmaMapperNode* pNode;
  DfLinuxDmaMapperContext* pMapperContext;
  DfDmaMapperPosition newPosition;
  struct page** pPages;
  struct page* p;
  size_t sliceIndex, slicePosition;
  size_t sliceOffset, sliceLength, sliceRemaining = 0;
  uint64_t sliceLocalAddress;
  size_t pageIndex, regionIndex;
  unsigned int pageOffset, pageRemaining;
  size_t chunk;
  unsigned int bufferPageOffset;
  unsigned int numEntry, maxEntry;
  dma_addr_t busAddress;

  dfAssert(pDevObj->hardwareId.busType == DfBusPci);
  dfAssert(numSlice > 0);
  dfAssert(mapIndex < pDevObj->system.dma.numMap);

  pDev = &pDevObj->system.device.pPci->dev;

  bufferPageOffset = (unsigned int)(uintptr_t)pBufferDescription->pAddress & pageOffsetMask;

  sliceIndex = pStartPosition->sliceIndex;
  dfAssert(sliceIndex < numSlice);
  slicePosition = pStartPosition->slicePosition;

  pMapperContext = &pDevObj->system.dma.pMapArray[mapIndex];
  maxEntry = pDevObj->system.dma.maxMapRegister;
  pNode = pMapperContext->pTable;

  numEntry = 0;
  while (sliceIndex < numSlice && numEntry < maxEntry) {
    /* Get current offset into current slice, and remaining bytes in current slice */
    pSlice = pSlices + sliceIndex;
    sliceOffset = pSlice->offset + slicePosition + bufferPageOffset;
    sliceLength = pSlice->length;
    dfAssert(slicePosition < sliceLength);
    sliceRemaining = sliceLength - slicePosition;
    sliceLocalAddress = pSlice->localAddress;
    if (!bFixedLocalAddress) {
      sliceLocalAddress += slicePosition;
    }
    /* Find corresponding page index in region and region index */
    pageOffset = (unsigned int)sliceOffset & pageOffsetMask;
    pageRemaining = PAGE_SIZE - pageOffset;
    pageIndex = sliceOffset >> PAGE_SHIFT;
    regionIndex = pageIndex / DF_BUFFER_DESCRIPTION_REGION_SIZE;
    pageIndex &= (DF_BUFFER_DESCRIPTION_REGION_SIZE - 1);
    /* Get corresponding page */
    pPages = pBufferDescription->system.region[regionIndex];
    p = pPages[pageIndex];
    /* Calculate how many bytes we can map this time around */
    chunk = (sliceRemaining > pageRemaining) ? pageRemaining : sliceRemaining;
    /* Map the page chunk */
    busAddress = dma_map_page(pDev, p, pageOffset, chunk, bWriteToDevice ? DMA_TO_DEVICE : DMA_FROM_DEVICE);
    pNode->busAddress = busAddress;
    pNode->length = chunk;
    pNode->localAddress = sliceLocalAddress;
    pNode++;
    /* Increment pointers etc. */
    slicePosition += chunk;
    if (slicePosition == sliceLength) {
      /* Reached end of current slice */
      sliceIndex++;
      slicePosition = 0;
    }
    numEntry++;
  }

  pMapperContext->curr.bWriteToDevice = bWriteToDevice;
  pMapperContext->curr.numEntry = numEntry;
  newPosition.sliceIndex = sliceIndex;
  newPosition.slicePosition = slicePosition;
  
  /* Finished, so call the client's callback function */
  pCallback(
    pDevObj,
    pMapperContext->index,
    pMapperContext->pTable,
    numEntry,
    &newPosition,
    pCallbackContext);

  return TRUE;
}

extern void
dfDmaMapperTeardown(
  DfDeviceObject* pDevObj,
  unsigned int mapIndex)
{
  struct device* pDev;
  DfDmaMapperNode* pNode;
  DfLinuxDmaMapperContext* pMapperContext;
  unsigned int i, numEntry;
  boolean_t bWriteToDevice;

  dfAssert(pDevObj->hardwareId.busType == DfBusPci);

  pDev = &pDevObj->system.device.pPci->dev;

  pMapperContext = &pDevObj->system.dma.pMapArray[mapIndex];
  pNode = pMapperContext->pTable;
  numEntry = pMapperContext->curr.numEntry;
  bWriteToDevice = pMapperContext->curr.bWriteToDevice;

  for (i = 0; i < numEntry; i++) {
    dma_unmap_page(pDev, pNode->busAddress, pNode->length, bWriteToDevice ? DMA_TO_DEVICE : DMA_FROM_DEVICE);
    pNode++;
  }
}
