/*
** File: event.c  
** Project: ADMXRC3 module driver
** Purpose: OS-independent IOCTL handlers for event registration and waiting.
**
** (C) Copyright Alpha Data 2009
*/

#include <df.h>
#include "device.h"
#include "event.h"

#if DF_HAS_USER_EVENT_HANDLE
static void
notifyAllEvents(
  Admxrc3DeviceContext* pDevCtx,
  unsigned int targetIndex)
{
  UserEventList* pList;
  UserEventInfo* pInfo;
  DfSpinLockFlags f;

  /* Signal all user events in the list */
  pList = &pDevCtx->fpga[targetIndex].interrupt.eventList;
  f = dfSpinLockGet(&pList->lock);
  pInfo = pList->pHead;
  while (NULL != pInfo) {
    dfUserEventSignal(pInfo->pUserEvent);
    pInfo = pInfo->pNext;
  }
  dfSpinLockPut(&pList->lock, f);
}
#endif

void
fpgaInterruptNotifyCallback(
  CoreNotification* pCoreNotification,
  unsigned int notifyType,
  void* pCallbackContext)
{
  FpgaInterruptNotification* pNotification = (FpgaInterruptNotification*)pCoreNotification;

  dfDebugPrint(0, ("fpgaInterruptNotifyCallback: pNotification=0x%p notifyType=%lu pCallbackContext=0x%p\n",
    pNotification, (unsigned long)notifyType, pCallbackContext));

#if DF_HAS_USER_EVENT_HANDLE
  notifyAllEvents((Admxrc3DeviceContext*)pCallbackContext, pNotification->targetIndex);
#endif
}

#if DF_HAS_USER_EVENT_HANDLE
/* Removes the user event list and frees all user event nodes that are associated with 'pClCtx' */
static void
cleanupUserEventList(
	Admxrc3DeviceContext* pDevCtx,
  Admxrc3ClientContext* pClCtx,
  UserEventList* pList)
{
  UserEventInfo* pInfo;
  UserEventInfo* pNext;
  DfSpinLockFlags f;

  f = dfSpinLockGet(&pList->lock);
  pInfo = pList->pHead;
  while (NULL != pInfo) {
    pNext = pInfo->pNext;
    if (pInfo->pClient == pClCtx) {
      if (NULL != pInfo->pNext) {
        pInfo->pNext->pPrev = pInfo->pPrev;
      } else {
        pList->pTail = pInfo->pNext;
      }
      if (NULL != pInfo->pPrev) {
        pInfo->pPrev->pNext = pInfo->pNext;
      } else {
        pList->pHead = pInfo->pNext;
      }
      dfUserEventPut(pInfo->pUserEvent);
      dfFree(pInfo);
    }
    pInfo = pNext;
  }
  dfSpinLockPut(&pList->lock, f);
}
#endif

#if DF_HAS_USER_EVENT_HANDLE
/* Removes from the device and frees all user event nodes that are associated with 'pClCtx' */
void
cleanupUserEvents(
	Admxrc3DeviceContext* pDevCtx,
  Admxrc3ClientContext* pClCtx)
{
  UserEventList* pList;
  unsigned int i;

  for (i = 0; i < MAX_NUM_TARGET_FPGA; i++) {
    pList = &pDevCtx->fpga[i].interrupt.eventList;
    cleanupUserEventList(pDevCtx, pClCtx, pList);
  }
}
#endif

#if DF_HAS_USER_EVENT_HANDLE
DfIoStatus
ioctlRegisterEvent(
	Admxrc3DeviceContext* pDevCtx,
  Admxrc3ClientContext* pClCtx,
  void* pBuffer,
  unsigned int inSize)
{
  IOCTLS_ADMXRC3_REGISTEREVENT* pIoctl = (IOCTLS_ADMXRC3_REGISTEREVENT*)pBuffer;
  DfIoStatus status = DfIoStatusSuccess;
  DfSpinLockFlags f;
  unsigned int notification, targetIndex;
  user_event_handle_t hEvent;
  DfUserEvent* pUserEvent;
  UserEventInfo* pInfo;
  UserEventList* pList;

  if (NULL == pIoctl || sizeof(pIoctl->in) != inSize) {
    return DfIoStatusInvalidCode;
  }

  notification = pIoctl->in.notification;
  hEvent = pIoctl->in.hEvent;

  /* Get the event list corresponding to 'notification' from the client context */
  if (notification >= ADMXRC3_EVENT_FPGAINTERRUPT(0) && notification <= ADMXRC3_EVENT_FPGAINTERRUPT(0xffU)) {
    targetIndex = notification - ADMXRC3_EVENT_FPGAINTERRUPT(0);
    if (targetIndex >= pDevCtx->coreInterface.pGetNumTargetFpga(pDevCtx->pCoreContext)) {
      return DfIoStatusError(ADMXRC3_INVALID_INDEX);
    }
    pList = &pDevCtx->fpga[targetIndex].interrupt.eventList;
  } else {
    return DfIoStatusError(ADMXRC3_INVALID_INDEX);
  }

  /* Convert the user-mode event handle to a kernel user-mode event */
  if (!dfUserEventGet(hEvent, &pUserEvent)) {
    return DfIoStatusError(ADMXRC3_INVALID_HANDLE);
  }

  /* Create a user-mode event information node */
  pInfo = (UserEventInfo*)dfMalloc(sizeof(UserEventInfo));
  if (NULL == pInfo) {
    dfUserEventPut(pUserEvent);
    return DfIoStatusError(ADMXRC3_NO_MEMORY);
  }
  pInfo->pClient = pClCtx;
  pInfo->hUserEvent = hEvent;
  pInfo->pUserEvent = pUserEvent;

  /* Add node to user-mode event list for client context */
  f = dfSpinLockGet(&pList->lock);
  pInfo->pNext = NULL;
  pInfo->pPrev = pList->pHead;
  if (NULL == pList->pHead) {
    pList->pHead = pInfo;
    pList->pTail = pInfo;
  } else {
    pList->pTail->pNext = pInfo;
    pList->pTail = pInfo;
  }
  dfSpinLockPut(&pList->lock, f);

  return status;
}
#endif

#if DF_HAS_USER_EVENT_HANDLE && DF_NEED_THUNK
DfIoStatus
ioctlRegisterEventThunk(
	Admxrc3DeviceContext* pDevCtx,
  Admxrc3ClientContext* pClCtx,
  void* pBuffer,
  unsigned int inSize)
{
  IOCTLS32_ADMXRC3_REGISTEREVENT* pIoctl32 = (IOCTLS32_ADMXRC3_REGISTEREVENT*)pBuffer;
  IOCTLS_ADMXRC3_REGISTEREVENT ioctl;

  if (NULL == pIoctl32 || sizeof(pIoctl32->in) != inSize) {
    return DfIoStatusInvalidCode;
  }

  ioctl.in.notification = pIoctl32->in.notification;
  ioctl.in.hEvent = dfThunkHandle(pIoctl32->in.hEvent);
  return ioctlRegisterEvent(pDevCtx, pClCtx, &ioctl, sizeof(ioctl.in));
}
#endif

#if DF_HAS_USER_EVENT_HANDLE
DfIoStatus
ioctlUnregisterEvent(
	Admxrc3DeviceContext* pDevCtx,
  Admxrc3ClientContext* pClCtx,
  void* pBuffer,
  unsigned int inSize)
{
  IOCTLS_ADMXRC3_UNREGISTEREVENT* pIoctl = (IOCTLS_ADMXRC3_UNREGISTEREVENT*)pBuffer;
  DfIoStatus status = DfIoStatusSuccess;
  DfSpinLockFlags f;
  unsigned int notification, targetIndex;
  user_event_handle_t hEvent;
  UserEventList* pList;
  UserEventInfo* pInfo;
  UserEventInfo* pNext;

  if (NULL == pIoctl || sizeof(pIoctl->in) != inSize) {
    return DfIoStatusInvalidCode;
  }

  notification = pIoctl->in.notification;
  hEvent = pIoctl->in.hEvent;

  /* Get the user-mode event list corresponding to 'notification' from the client context */
  if (notification >= ADMXRC3_EVENT_FPGAINTERRUPT(0) && notification <= ADMXRC3_EVENT_FPGAINTERRUPT(0xffU)) {
    targetIndex = notification - ADMXRC3_EVENT_FPGAINTERRUPT(0);
    if (targetIndex >= pDevCtx->coreInterface.pGetNumTargetFpga(pDevCtx->pCoreContext)) {
      return DfIoStatusError(ADMXRC3_INVALID_INDEX);
    }
    pList = &pDevCtx->fpga[targetIndex].interrupt.eventList;
  } else {
    return DfIoStatusError(ADMXRC3_INVALID_INDEX);
  }

  /* Search the user-mode event list for a node matching 'hEvent', and remove it from the list if found */
  f = dfSpinLockGet(&pList->lock);
  pInfo = pList->pHead;
  while (NULL != pInfo) {
    pNext = pInfo->pNext;
    if (pInfo->pClient == pClCtx && pInfo->hUserEvent == hEvent) {
      /* Found the node; remove it from the event list */
      if (NULL != pInfo->pNext) {
        pInfo->pNext->pPrev = pInfo->pPrev;
      } else {
        pList->pTail = pInfo->pPrev;
      }
      if (NULL != pInfo->pPrev) {
        pInfo->pPrev->pNext = pInfo->pNext;
      } else {
        pList->pHead = pInfo->pNext;
      }
      break;
    }
    pInfo = pNext;
  }
  dfSpinLockPut(&pList->lock, f);

  /* Nothing was found in the list; hence 'hEvent' must be considered invalid */
  if (NULL == pInfo) {
    return DfIoStatusError(ADMXRC3_INVALID_HANDLE);
  }

  /* Found a match; release the kernel user-mode event & free the list node */
  dfUserEventPut(pInfo->pUserEvent);
  dfFree(pInfo);

  return status;
}
#endif

#if DF_HAS_USER_EVENT_HANDLE && DF_NEED_THUNK
DfIoStatus
ioctlUnregisterEventThunk(
	Admxrc3DeviceContext* pDevCtx,
  Admxrc3ClientContext* pClCtx,
  void* pBuffer,
  unsigned int inSize)
{
  IOCTLS_ADMXRC3_UNREGISTEREVENT* pIoctl32 = (IOCTLS_ADMXRC3_UNREGISTEREVENT*)pBuffer;
  IOCTLS_ADMXRC3_UNREGISTEREVENT ioctl;

  if (NULL == pIoctl32 || sizeof(pIoctl32->in) != inSize) {
    return DfIoStatusInvalidCode;
  }

  ioctl.in.notification = pIoctl32->in.notification;
  ioctl.in.hEvent = dfThunkHandle(pIoctl32->in.hEvent);
  return ioctlUnregisterEvent(pDevCtx, pClCtx, &ioctl, sizeof(ioctl.in));
}
#endif
