/*
** File: info.c  
** Project: ADMXRC3 module driver
** Purpose: OS-independent IOCTL handlers for returning information about a
**          card. None of these IOCTL handlers modifies the state of a card,
**          so they can be considered "read-only".
**
** (C) Copyright Alpha Data 2009
*/

#include <df.h>
#include "info.h"
#include <admxrc3/status.h>
#include <admxrc3/types.h>

void
setupIoctlsGetFpgaInfo(
	Admxrc3DeviceContext* pDevCtx,
  IOCTLS_ADMXRC3_GETFPGAINFO* pIoctl,
  ADMXRC3_FPGA_TYPE device,
  ADMXRC3_PACKAGE_TYPE package,
  const char* pSpeedGrade,
  const char* pStepping,
  boolean_t bPresent)
{
  pIoctl->out.deviceCode = device;
  pIoctl->out.packageCode = package;
  if (device <= 3) {
    pIoctl->out.familyCode = ADMXRC3_FAMILY_4K;
    pIoctl->out.subfamilyCode = ADMXRC3_SUBFAMILY_NONE;
  } else if (device <= 31) {
    pIoctl->out.familyCode = ADMXRC3_FAMILY_VIRTEX;
    if (device <= 7) {
      pIoctl->out.subfamilyCode = ADMXRC3_SUBFAMILY_NONE;
    } else if (device <= 11) {
      pIoctl->out.subfamilyCode = ADMXRC3_SUBFAMILY_E;
    } else {
      pIoctl->out.subfamilyCode = ADMXRC3_SUBFAMILY_EM;
    }
  } else if (device <= 63) {
    pIoctl->out.familyCode = ADMXRC3_FAMILY_VIRTEX2;
    pIoctl->out.subfamilyCode = ADMXRC3_SUBFAMILY_NONE;
  } else if (device <= 95) {
    pIoctl->out.familyCode = ADMXRC3_FAMILY_VIRTEX2P;
    if (device <= 79) {
      pIoctl->out.subfamilyCode = ADMXRC3_SUBFAMILY_NONE;
    } else {
      pIoctl->out.subfamilyCode = ADMXRC3_SUBFAMILY_2PROX;
    }
  } else if (device <= 127) {
    pIoctl->out.familyCode = ADMXRC3_FAMILY_VIRTEX4;
    if (device <= 103) {
      pIoctl->out.subfamilyCode = ADMXRC3_SUBFAMILY_4LX;
    } else if (device <= 111) {
      pIoctl->out.subfamilyCode = ADMXRC3_SUBFAMILY_4SX;
    } else {
      pIoctl->out.subfamilyCode = ADMXRC3_SUBFAMILY_4FX;
    }
  } else if (device <= 159) {
    pIoctl->out.familyCode = ADMXRC3_FAMILY_VIRTEX5;
    if (device <= 135) {
      pIoctl->out.subfamilyCode = ADMXRC3_SUBFAMILY_5LX;
    } else if (device <= 143) {
      pIoctl->out.subfamilyCode = ADMXRC3_SUBFAMILY_5LXT;
    } else if (device <= 151) {
      pIoctl->out.subfamilyCode = ADMXRC3_SUBFAMILY_5SXT;
    } else {
      pIoctl->out.subfamilyCode = ADMXRC3_SUBFAMILY_5FXT;
    }
  } else if (device <= 191) {
    pIoctl->out.familyCode = ADMXRC3_FAMILY_VIRTEX6;
    if (device <= 167) {
      pIoctl->out.subfamilyCode = ADMXRC3_SUBFAMILY_6LX;
    } else if (device <= 175) {
      pIoctl->out.subfamilyCode = ADMXRC3_SUBFAMILY_6LXT;
    } else if (device <= 183) {
      pIoctl->out.subfamilyCode = ADMXRC3_SUBFAMILY_6SXT;
    } else {
      pIoctl->out.subfamilyCode = ADMXRC3_SUBFAMILY_6HXT;
    }
  } else {
    pIoctl->out.familyCode = (ADMXRC3_FAMILY_TYPE)-1;
    pIoctl->out.subfamilyCode = ADMXRC3_SUBFAMILY_NONE;
  }
  memset(pIoctl->out.speedGrade, 0, sizeof(pIoctl->out.speedGrade));
  if (NULL != pSpeedGrade) {
    dfStrCpy(pIoctl->out.speedGrade, sizeof(pIoctl->out.speedGrade) / sizeof(pIoctl->out.speedGrade[0]), pSpeedGrade);
  }
  memset(pIoctl->out.stepping, 0, sizeof(pIoctl->out.stepping));
  if (NULL != pStepping) {
    dfStrCpy(pIoctl->out.stepping, sizeof(pIoctl->out.stepping) / sizeof(pIoctl->out.stepping[0]), pStepping);
  }
  pIoctl->out.bPresent = bPresent;
}

DfIoStatus
ioctlGetBankInfo(
	Admxrc3DeviceContext* pDevCtx,
    void* pBuffer,
    unsigned int inSize,
    unsigned int outSize)
{
    IOCTLS_ADMXRC3_GETBANKINFO* pIoctl = (IOCTLS_ADMXRC3_GETBANKINFO*)pBuffer;
    DfIoStatus status = DfIoStatusSuccess;
    ADMXRC3_STATUS methodStatus;
    const void* pVpd;
    unsigned int bankIndex;

    if (NULL == pIoctl || sizeof(pIoctl->in) != inSize || sizeof(pIoctl->out) != outSize) {
        return DfIoStatusInvalidCode;
    }

    bankIndex = pIoctl->in.bankIndex;

    pVpd = pDevCtx->coreInterface.pGetVpdPointer(pDevCtx->pCoreContext);
    if (NULL == pVpd) {
        /* No VPD, return card info structure that is zeroes except for model field */
        dfDebugPrint(0, ("+++ ioctlGetBankInfo: no VPD\n"));
        status = DfIoStatusError(ADMXRC3_HARDWARE_ERROR);
        return status;
    }

    memset(&pIoctl->out, 0, sizeof(pIoctl->out));
    methodStatus = pDevCtx->methods.pGetBankInfo(pDevCtx, bankIndex, pVpd, pIoctl);
    if (ADMXRC3_SUCCESS != methodStatus) {
        status = DfIoStatusError(methodStatus);
    }
    return status;
}

DfIoStatus
ioctlGetCardInfo(
	Admxrc3DeviceContext* pDevCtx,
  void* pBuffer,
  unsigned int outSize)
{
  IOCTLS_ADMXRC3_GETCARDINFO* pIoctl = (IOCTLS_ADMXRC3_GETCARDINFO*)pBuffer;
  DfIoStatus status = DfIoStatusSuccess;
  const void* pVpd;
  CoreModelType model;

  if (NULL == pIoctl || sizeof(pIoctl->out) != outSize) {
    return DfIoStatusInvalidCode;
  }

  memset(&pIoctl->out, 0, sizeof(pIoctl->out));

  model = pDevCtx->coreInterface.pGetModel(pDevCtx->pCoreContext);
  pIoctl->out.model = model & 0xfffU;
  pIoctl->out.numClockGen = pDevCtx->coreInterface.pGetNumClockGen(pDevCtx->pCoreContext);
  pIoctl->out.numDmaChannel = pDevCtx->coreInterface.pGetNumDmaChannel(pDevCtx->pCoreContext);
  pIoctl->out.numFlashBank = pDevCtx->coreInterface.pGetNumFlashBank(pDevCtx->pCoreContext);
  pIoctl->out.numTargetFpga = pDevCtx->coreInterface.pGetNumTargetFpga(pDevCtx->pCoreContext);
  pIoctl->out.numWindow = pDevCtx->coreInterface.pGetNumWindow(pDevCtx->pCoreContext);

  pVpd = pDevCtx->coreInterface.pGetVpdPointer(pDevCtx->pCoreContext);
  if (NULL == pVpd) {
    /* No VPD, return card info structure that is zeroes except for model field */
    dfDebugPrint(3, ("+++ ioctlGetCardInfo: no VPD\n"));
  } else {
    pDevCtx->methods.pGetCardInfo(pDevCtx, pVpd, pIoctl);
  }

  return status;
}

DfIoStatus
ioctlGetFpgaInfo(
	Admxrc3DeviceContext* pDevCtx,
  void* pBuffer,
  unsigned int inSize,
  unsigned int outSize)
{
  IOCTLS_ADMXRC3_GETFPGAINFO* pIoctl = (IOCTLS_ADMXRC3_GETFPGAINFO*)pBuffer;
  DfIoStatus status = DfIoStatusSuccess;
  ADMXRC3_STATUS methodStatus;
  const void* pVpd;
  unsigned int targetIndex;

  if (NULL == pIoctl || sizeof(pIoctl->in) != inSize || sizeof(pIoctl->out) != outSize) {
    return DfIoStatusInvalidCode;
  }

  targetIndex = pIoctl->in.targetIndex;

  pVpd = pDevCtx->coreInterface.pGetVpdPointer(pDevCtx->pCoreContext);
  if (NULL == pVpd) {
    dfDebugPrint(0, ("+++ ioctlGetFpgaInfo: no VPD\n"));
    status = DfIoStatusError(ADMXRC3_HARDWARE_ERROR);
    return status;
  }

  memset(&pIoctl->out, 0, sizeof(pIoctl->out));
  methodStatus = pDevCtx->methods.pGetFpgaInfo(pDevCtx, targetIndex, pVpd, pIoctl);
  if (ADMXRC3_SUCCESS != methodStatus) {
    status = DfIoStatusError(methodStatus);
  }

  return status;
}

DfIoStatus
ioctlGetDriverVersion(
	Admxrc3DeviceContext* pDevCtx,
  void* pBuffer,
  unsigned int outSize)
{
  IOCTLS_ADMXRC3_GETDRIVERVERSION* pIoctl = (IOCTLS_ADMXRC3_GETDRIVERVERSION*)pBuffer;
  DfIoStatus status = DfIoStatusSuccess;

  if (NULL == pIoctl || sizeof(pIoctl->out) != outSize) {
    return DfIoStatusInvalidCode;
  }

  pIoctl->out.major = ADB3_VERSION_0;
  pIoctl->out.minor = ADB3_VERSION_1;
  pIoctl->out.bugfix = ADB3_VERSION_2;

  return status;
}

DfIoStatus
ioctlGetWindowInfo(
	Admxrc3DeviceContext* pDevCtx,
  void* pBuffer,
  unsigned int inSize,
  unsigned int outSize)
{
  IOCTLS_ADMXRC3_GETWINDOWINFO* pIoctl = (IOCTLS_ADMXRC3_GETWINDOWINFO*)pBuffer;
  DfIoStatus status = DfIoStatusSuccess;
  CoreWindowStatus windowStatus;
  CoreWindowInfo windowInfo;
  unsigned int windowIndex;

  if (NULL == pIoctl || sizeof(pIoctl->in) != inSize || sizeof(pIoctl->out) != outSize) {
    return DfIoStatusInvalidCode;
  }

  windowIndex = pIoctl->in.windowIndex;
  windowStatus = pDevCtx->coreInterface.pGetWindowInfo(pDevCtx->pCoreContext, windowIndex, &windowInfo);
  if (CoreWindowSuccess != windowStatus) {
    switch (windowStatus) {
    case CoreWindowInvalidIndex:
      status = DfIoStatusError(ADMXRC3_INVALID_INDEX);
      break;

    default:
      status = DfIoStatusError(ADMXRC3_UNEXPECTED_ERROR);
      break;
    }
  } else {
    memset(&pIoctl->out, 0, sizeof(pIoctl->out));
    pIoctl->out.busBase = windowInfo.busBase;
    pIoctl->out.busSize = windowInfo.busSize;
    pIoctl->out.localSize = windowInfo.localSize;
    pIoctl->out.localBase = windowInfo.localBase;
    pIoctl->out.virtualSize = (uint64_t)windowInfo.kernelSize;
  }
  return status;
}
